//! Handles building the database, and exposes data structures for the database and ways to interact with them
use super::*;
use std::{
    collections::{
	HashMap,
	HashSet,
    },
    path::{
	PathBuf,
	Path,
    },
    cmp::{
	PartialOrd,
	Ordering,
    },
    marker::{
	Unpin,
    },
};
use chrono::{
    DateTime,
    Utc,
};
use tokio::{
    prelude::*,
    io::{
	AsyncRead,
    },
};

#[derive(Debug, PartialEq, Eq)]
pub struct Database
{
    /// Key here is the actual location in /tmp/videl of the database and versioned files
    database: HashMap<PathBuf,Entry>,
}

/// All versions of a single file
#[derive(Debug, PartialEq, Eq)]
pub struct Entry
{
    /// Versions are a hashset because we don't need to version files that are /entirely/ identical. (i.e. same timestamp and hash)
    versions: HashSet<Version>,
}

/// A single version of a single file
#[derive(Debug, PartialEq, Eq, Hash)]
pub struct Version
{
    date: DateTime<Utc>,
    hash: hash::Sha256Sum,
    size: usize,
}

impl PartialOrd for Version
{
    #[inline] fn partial_cmp(&self, other: &Self) -> Option<Ordering>
    {
	self.date.partial_cmp(&other.date)
    }
}

impl Version
{
    /// Gets the lowermost name of the path for this specific version
    pub fn get_pathname_string(&self) -> String
    {
	format!("{}", self.date.timestamp_nanos())
    }

    /// Create a new `Version` from a stream source with a specific UTC timestamp.
    pub async fn new_specific<R>(timestamp: DateTime<Utc>, from: &mut R) -> io::Result<Self>
	where R: AsyncRead + Unpin
    {
	let (size, hash) = {
	    let mut hash = hash::Sha256Sum::empty();
	    (hash.compute_into(from).await?, hash)
	};

	Ok(Self {
	    date: timestamp,
	    hash,
	    size,
	})
    }

    /// Create a new `Version` from a stream sorce with the current UTC timestamp.
    pub async fn new<R>(from: &mut R) -> io::Result<Self>
	where R: AsyncRead + Unpin
    {
	Self::new_specific(Utc::now(), from).await
    }

    /// Length of this `Version`
    #[inline] pub fn len(&self) -> usize
    {
	self.size
    }

    /// Timestamp of this `Version`
    #[inline] pub fn timestamp(&self) -> &DateTime<Utc>
    {
	&self.date
    }

    /// Hash of this `Version`
    #[inline] pub fn hash(&self) -> &hash::Sha256Sum
    {
	&self.hash
    }

    //TODO: How do we manage serialisation? Do we save all this metadata, or just recalculate it from the filename?
}
