//! Error handling stuffs (I think we'll go with this one in the end)
use std::{
    error,
    io,
    cmp,
    fmt,
    str::{
	FromStr,
    },
    borrow::Cow,
    marker::{
	Send, Sync
    },
};

/// Result for all our operations, see `ErrorKind` below
///
/// This type can be used for when we want to propagate an error, but not pass it off to `eyre` reporter just yet, because we might still want to act on it and respond.
pub type Result<T> = std::result::Result<T, Error>;

mod context;
pub use context::*;
mod message;
use message::*;
mod wrap;
pub use wrap::*;
mod aggregate;
pub use aggregate::*;

impl Default for ErrorKind
{
    #[inline]
    fn default() -> Self
    {
	Self::None
    }
}

impl Error
{
    /// Into the `ErrorKind`
    pub fn into_inner(self) -> ErrorKind
    {
	*self.internal
    }
    
    /// The kind of the error
    pub fn kind(&self) -> &ErrorKind
    {
	&self.internal
    }

    /// The message for this error
    pub fn message(&self) -> Option<Cow<'_, str>>
    {
	self.message.as_str()
    }

    /// The context from this error
    pub fn context(&self) -> &Context
    {
	&self.context
    }

    /// Create a new error with a constant string
    pub fn new_static(err: ErrorKind, ctx: Context, message: &'static str) -> Self
    {
	Self {
	    internal: box err,
	    context: ctx,
	    message: Message::from_str(message),
	}
    }
    /// Create a new error with a string
    pub fn new_string(err: ErrorKind, ctx: Context, message: String) -> Self
    {
	Self {
	    internal: box err,
	    context: ctx,
	    message: Message::from_string(message),
	}
    }
    /// Create a new error with a message
    pub fn new(err: ErrorKind, ctx: Context, message: impl fmt::Display + 'static + Send + Sync) -> Self
    {
	Self {
	    internal: box err,
	    context: ctx,
	    message: Message::from_other(message),
	}
    }
    /// Create a new error with no message
    pub fn new_none(err: ErrorKind, ctx: Context) -> Self
    {
	Self {
	    internal: box err,
	    context: ctx,
	    message: Message::None,
	}
    }
}

impl fmt::Display for Error
{
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result
    {
	if !self.context().is_empty() {
	    write!(f, "{}: ", self.context)?;
	}
	match &self.message {
	    Message::None => write!(f, "{}", self.internal),
	    message => write!(f, "{}", message.fmt())
	}
    }
}
impl error::Error for Error
{
    fn source(&self) -> Option<&(dyn error::Error + 'static)>
    {
	match &self.message {
	    Message::None => std::iter::successors(self.internal.source(), |e| e.source()).take(1).next(),
	    _ => Some(&self.internal),
	}
    }
}

#[macro_export] macro_rules! context {
    () => ($crate::error::Context::with_all(file!(), line!(), column!()));
    ($msg:expr) => ($crate::error::ContextMessage::new($crate::context!(), $msg));
    (yield $fmt:literal $($tt:tt)*) => ($crate::context!(lazy_format::lazy_format!($fmt $($tt)*)));
    ($fmt:literal $($tt:tt)*) => ($crate::context!(format!($fmt $($tt)*)));
}

/// Macro for instantiating errors and also ad-hoc error types
///
/// # Instantiating `Error`.
/// The primary use is to return errors of the `Error` struct, this is how it can work.
/// ## With `ErrorKind`
/// The kind should be prefixed with `;`:
/// ```
/// error!(ErrorKind::Unknown; "unknown error message");
/// error!(ErrorKind::IO(yield io_err; "I/O error opening {:?}", file));
/// ```
///
/// ## Without `ErrorKind`
/// With all except a single string literal, the message is boxed as the `ErrorKind`, and printed with a `None` message.
/// This means that the message given as an ad-hoc error type is counted as the source.
///
/// ```
/// error!("unknown error");
/// error!(yield "something went wrong: {}", message);
/// ```
///
/// # Defining errors
/// To create ad-hoc errors, use `error!(impl ...)`. It has the same syntax as the above cases, except with no error kind
/// ```
/// error!(impl "Error message");
/// error!(impl yield "lazy formatted string {}", string);
/// ```
///
/// # `yield`ing formatting
/// When format string are provided to `error!()`, they by default will use the eager `format!()` macro, which will immediately evaluate it's arguments and produce the string to put into the error.
/// To instead box the arguments and evaluate as-needed, we can use the `lazy_format` crate by prefixing the format string literals with `yield`. This works on both `Error` instantiations and ad-hoc error definitions.
#[macro_export] macro_rules! error {
    (@ $msg:expr) => {
	{
	    pub struct DisplayError<D>(D) where D: fmt::Display;
	    impl<D> fmt::Debug for DisplayError<D>
	    where D: fmt::Display
	    {
		#[inline] fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result
		{
		    write!(f, "{}", std::any::type_name::<Self>())
		}
	    }
	    impl<D> fmt::Display for DisplayError<D>
	    where D: fmt::Display
	    {
		#[inline] fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result
		{
		    self.0.fmt(f)
		}
	    }
	    impl<D> error::Error for DisplayError<D>
	    where D: fmt::Display{}

	    DisplayError($msg)
	}
    };

    (impl $lit:literal) => ($crate::error!(@ $lit));
    (impl $expr:expr) => ($crate::error!(@ $expr));
    (impl yield $lit:literal $($tt:tt)*) => ($crate::error!(@ lazy_format::lazy_format!($lit $($tt)*)));
    (impl $lit:literal $($tt:tt)*) => ($crate::error!(@ format!($lit $($tt)*)));
    
    ($kind:expr; $lit:literal) => ($crate::error::Error::new_static($kind, $crate::context!(), $lit));
    ($lit:literal) => ($crate::error!(Default::default(), $lit));
    
    ($kind:expr; $msg:expr) => ($crate::error::Error::new($kind, $crate::context!(), $msg));
    ($msg:expr) => ($crate::error::Error::new_none($crate::error::ErrorKind::Any(box $crate::error!(@ $msg)), $crate::context!()));

    (yield $lit:literal $($tt:tt)*) => ($crate::error::Error::new_none($crate::error::ErrorKind::Any(box $crate::error!(@ lazy_format::lazy_format!($lit $($tt)*))), $crate::context!()));
    ($lit:literal $($tt:tt)*) => ($crate::error::Error::new_none($crate::error::ErrorKind::Any(box $crate::error!(@ format!($lit $($tt)*))), $crate::context!()));
    
    (yield $kind:expr; $lit:literal $($tt:tt)*) => ($crate::error!($kind; lazy_format::lazy_format!($lit $($tt)*)));
    ($kind:expr; $lit:literal $($tt:tt)*) => ($crate::error::Error::new_string($kind, $crate::context!(), format!($lit $($tt)*)));
}

#[derive(Debug)]
pub struct Error
{
    internal: Box<ErrorKind>,
    context: Context,
    message: Message,
}

// Actual stuffs

#[non_exhaustive]
#[derive(Debug)]
pub enum ErrorKind
{
    IO(io::Error),
    Wrap(Error),
    Any(Box<dyn error::Error +Send + Sync + 'static>),
    None,
    Unknown,
}

impl fmt::Display for ErrorKind
{
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result
    {
	match self {
	    Self::IO(io) => write!(f, "i/o error: {}", io),
	    Self::Wrap(wrap) => write!(f, "{}", wrap),
	    Self::Any(any) => write!(f, "{}", any),
	    Self::None => Ok(()),
	    _ => write!(f, "unknown error"),
	}
    }
}

impl error::Error for ErrorKind
{
    fn source(&self) -> Option<&(dyn error::Error + 'static)>
    {
	Some(match &self {
 	    Self::IO(io) => io,
	    Self::Wrap(wrap) => return wrap.source(),
	    Self::Any(any) => any.as_ref(),
	    _ => return None,
	})
    }
}

impl From<io::Error> for ErrorKind
{
    fn from(from: io::Error) -> Self
    {
	Self::IO(from)
    }
}

impl<T> From<T> for Error
where T: Into<ErrorKind>
{
    fn from(from: T) -> Self
    {
	Self {
	    internal: box from.into(),
	    context: Default::default(),
	    message: Message::None,
	}
    }
}

#[cfg(test)]
#[allow(unused_imports)]
mod test
{
    use super::*;
    use color_eyre::{
	eyre::{
	    Result as ReportedResult,
	},
	Help,
	SectionExt,
    };
    #[test]
    fn test() -> ReportedResult<()>
    {
	fn testt() -> Result<()> {
	    let err= error!(yield one().unwrap_err().into_inner(); "hello world {} {}", 1, 2);
	    Err(err)
		.wrap_err_with(|| context!("with string"))
		.wrap_err_with(|| context!(yield "with another string and some stuffs {}", 1))?;
	    Ok(())
	}

	fn works() -> Result<()>
	{
	    Ok(())
	}

	fn one() -> Result<()>
	{
	    Err(std::io::Error::new(std::io::ErrorKind::UnexpectedEof, "end of records"))?;
	    Ok(())
	}
	
	fn two() -> Result<()>
	{
	    Err(std::io::Error::new(std::io::ErrorKind::NotFound, "nonexistent"))?;
	    Ok(())
	}
	
	fn three() -> Result<()>
	{
	    Err(std::io::Error::new(std::io::ErrorKind::BrokenPipe, "piped"))
		.wrap_err_with(|| context!("hello hello"))?;
	    Ok(())
	}

	

	Ok(one()
	   .aggregate_with(|_| testt())
	   .aggregate_with(|_| two())
	   .aggregate_with(|_| works())
	   .aggregate_with(|_| three())?)
    }

    #[test]
    fn strings() -> ReportedResult<()>
    {
	Err(error!(yield "literal format {}", 1))?;
	Ok(())
    }
}
